package java.util.concurrent;

import java.util.Spliterators;
import java.util.Spliterator;
import java.util.function.Consumer;
import java.util.function.Predicate;
import java.util.Set;
import java.util.Iterator;
import java.util.Collection;
import java.io.Serializable;
import java.util.AbstractSet;

public class CopyOnWriteArraySet<E> extends AbstractSet<E> implements Serializable
{
    private static final long serialVersionUID = 5457747651344034263L;
    private final CopyOnWriteArrayList<E> al;
    
    public CopyOnWriteArraySet() {
        this.al = new CopyOnWriteArrayList();
    }
    
    public CopyOnWriteArraySet(Collection<? extends E> collection) {
        if (collection.getClass() == CopyOnWriteArraySet.class) {
            this.al = new CopyOnWriteArrayList(((CopyOnWriteArraySet)collection).al);
        }
        else {
            (this.al = new CopyOnWriteArrayList()).addAllAbsent(collection);
        }
    }
    
    @Override
    public int size() {
        return this.al.size();
    }
    
    @Override
    public boolean isEmpty() {
        return this.al.isEmpty();
    }
    
    @Override
    public boolean contains(Object o) {
        return this.al.contains(o);
    }
    
    @Override
    public Object[] toArray() {
        return this.al.toArray();
    }
    
    @Override
    public <T> T[] toArray(T[] array) {
        return this.al.toArray(array);
    }
    
    @Override
    public void clear() {
        this.al.clear();
    }
    
    @Override
    public boolean remove(Object o) {
        return this.al.remove(o);
    }
    
    @Override
    public boolean add(E e) {
        return this.al.addIfAbsent(e);
    }
    
    @Override
    public boolean containsAll(Collection<?> collection) {
        return this.al.containsAll(collection);
    }
    
    @Override
    public boolean addAll(Collection<? extends E> collection) {
        return this.al.addAllAbsent(collection) > 0;
    }
    
    @Override
    public boolean removeAll(Collection<?> collection) {
        return this.al.removeAll(collection);
    }
    
    @Override
    public boolean retainAll(Collection<?> collection) {
        return this.al.retainAll(collection);
    }
    
    @Override
    public Iterator<E> iterator() {
        return this.al.iterator();
    }
    
    @Override
    public boolean equals(Object o) {
        if (o == this) {
            return true;
        }
        if (!(o instanceof Set)) {
            return false;
        }
        Iterator<Object> iterator = ((Set)(Set)o).iterator();
        Object[] array = this.al.getArray();
        int length = array.length;
        boolean[] array2 = new boolean[length];
        int n = 0;
    Label_0054:
        while (iterator.hasNext()) {
            if (++n > length) {
                return false;
            }
            Object next = iterator.next();
            int i = 0;
            while (i < length) {
                if (!array2[i] && eq(next, array[i])) {
                    array2[i] = true;
                    continue Label_0054;
                }
                ++i;
            }
            return false;
        }
        return n == length;
    }
    
    @Override
    public boolean removeIf(Predicate<? super E> predicate) {
        return this.al.removeIf(predicate);
    }
    
    @Override
    public void forEach(Consumer<? super E> consumer) {
        this.al.forEach(consumer);
    }
    
    @Override
    public Spliterator<E> spliterator() {
        return Spliterators.spliterator(this.al.getArray(), 1025);
    }
    
    private static boolean eq(Object o, Object obj) {
        return (o == null) ? (obj == null) : o.equals(obj);
    }
}
